#include "fileCache.h"
#include <cstdio>
#include "stdint.h"

CFileCache   fileCache;

//-----------------------------------------------------------------------------------
CFileData::CFileData()
{
    mem = NULL;
    size = 0;
}

//-----------------------------------------------------------------------------------
CFileCache::CFileCache()
{
}

//-----------------------------------------------------------------------------------
CFileCache::~CFileCache()
{
    std::map<std::string, CFileData>::iterator it;

    for (it = mCache.begin(); it != mCache.end(); ++it)
    {
        if (it->second.mem != NULL)
            delete[] (uint8_t*)it->second.mem;
    }

    mCache.clear();
    mImageCache.clear();
}

//-----------------------------------------------------------------------------------
CFileData   CFileCache::getFile(std::string fileName)
{
    std::map<std::string, CFileData>::iterator it = mCache.find(fileName);

    if (it != mCache.end())
        return it->second;

    CFileData   data;
    FILE*   handle = fopen(fileName.c_str(), "rb");

    if (handle != NULL)
    {
        fseek(handle, 0, SEEK_END);
        data.size = ftell(handle);
        fseek(handle, 0, SEEK_SET);
        data.mem = (void*)(new uint8_t[data.size]);

        if (data.mem != NULL)
            if (fread(data.mem, 1, data.size, handle) != data.size)
            {
                delete[] (uint8_t*)data.mem;
                data.mem = NULL;
            }

        fclose(handle);

        if (data.mem != NULL)
            mCache[fileName] = data;
    }
    return data;
}

//-----------------------------------------------------------------------------------
QImage      CFileCache::getImage(std::string fileName)
{
    std::map<std::string, QImage>::iterator it = mImageCache.find(fileName);

    if (it != mImageCache.end())
        return it->second;

    static QImage   image;
    CFileData   data = getFile(fileName);

    if (data.mem != NULL)
    {
        image.loadFromData((const uchar *)data.mem, data.size);
        mImageCache[fileName] = image;
    }

    return image;
}
